@@ -20,7 +20,12 @@ static inline bool
 can_strip_from_end_of_url(uint32_t ch) {
     // remove trailing punctuation
     return (
-        (is_P_category(ch) && ch != '/') ||
+        (is_P_category(ch) && ch != '/' && ch != '&' && ch != '-') ||
         ch == '>'
     ) ? true : false;
 }
+
+static inline bool
+is_private_use(char_type ch) {
+    return (0xe000 <= ch && ch <= 0xf8ff) || (0xF0000 <= ch && ch <= 0xFFFFF) || (0x100000 <= ch && ch <= 0x10FFFF);
+}
